import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "moebius92.Misc";
import "moebius92.RaceTrack";

recovery = { blinding = true, bane = true, cracked = false, herb = false, storm = true, march = true, light = true, test = true };

quickslot = { blinding = "0x70003152", bane = "0x7000B60D", cracked = "0x70003143", herb = "0x70003EC0",
              storm = "0x7000B63C", march = "0x70000FA7", light = "0x70003EC2", test = "0x70003EB9" };

ccName = {
	["Bane Flare"] = "bane",
	["Blinding Flash"] = "blinding",
	["Cracked Earth"] = "cracked",
	["Ents go to War"] = "march",
	["Herb-lore"] = "herb",
	["Light of the Rising Dawn"] = "light",
	["Storm-lore"] = "storm",
	["Test of Will"] = "test"
};

CCTimerRegime = class(Turbine.UI.Control);

function CCTimerRegime:Constructor(msg)
	Turbine.UI.Control.Constructor(self);
	self.msg = msg;
	self:SetMouseVisible(false);
end

function CCTimerRegime:MouseEnter(args)
	Turbine.Shell.WriteLine(self.msg);
end

function CCTimerRegime:MouseLeave(args)
	Turbine.Shell.WriteLine("---");
end

CCTimerWindow = class(Turbine.UI.Window);

function CCTimerWindow:Constructor()
	Turbine.UI.Window.Constructor(self);

	self:LoadSettings();

	self.hide = false;
	self.visible = false;
	self.size = 36;
	self.seconds = 30;
	self.dragging = false;
	self.move = false;
	self.lanes = 1;
	self:SetPosition(self.settings.x, self.settings.y);
	self:SetMouseVisible(false);
	self.lastTime = Turbine.Engine.GetGameTime();
	self.track = moebius92.RaceTrack.RaceTrack(self.size, 0);
	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.name = self.player:GetName();
	self.duration = { };
	self.cc = { };
	self.preroot = 0;

	Turbine.PluginManager.LoadPlugin("SkillInfo");

	self.mark = { };
	for k, v in pairs(ccName) do
		self.mark[v] = 0;
	end

	self:SetBackColor(Turbine.UI.Color(0.25, 0.0, 1.0, 0.0));
	self:SetWantsKeyEvents(true);
	skillInfoWindow:AddQuickslotHandler(function(skill, time, t)
		local n = ccName[skill];

		if n ~= nil then
			if t == "P" then
				self.mark[n] = time;
			elseif type(t) == "table" and t.H ~= nil then
				if (time - self.mark[n]) < 2.0 then
					self:AddCC(n, self.mark[n]);
				end
				self.mark[n] = 0;
			end
		elseif skill == "!update" then
			self:GetTimes();
		end
	end);
	self:GetTimes();
	self:Layout(true);
	self:SetVisible();
end

function CCTimerWindow:GetTimes()
	local s;

	self.preroot = 10;
	self.duration["bane"] = 15;
	self.duration["blinding"] = 30;
	self.duration["cracked"] = 30;
	self.duration["herb"] = 30;
	self.duration["light"] = 3;
	self.duration["march"] = 6;
	self.duration["test"] = 5;
	self.duration["storm"] = 3;

	s = skillInfoWindow:getTraitSet();
	if s["R"] >= 3 then
		self.duration["blinding"] = self.duration["blinding"] - 10;
	end
	if s["R"] >= 4 then
		self.duration["blinding"] = self.duration["blinding"] - 15;
	end
	if s["Y"] >= 3 then
		self.preroot = self.preroot - 5;
	end

	s = skillInfoWindow:getEquipSet();
	if s["OrthancY"] ~= nil and s["OrthancY"] >= 5 then
		self.duration["blinding"] = self.duration["blinding"] + 5;
	end
	if s["BG"] ~= nil and s["BG"] >= 5 then
		self.duration["blinding"] = self.duration["blinding"] + 5;
	end
end

function CCTimerWindow:SetVisible(v)
	if v ~= nil then
		self.visible = v;
	end

	Turbine.UI.Control.SetVisible(self, self.visible and not self.hide);
end

function CCTimerWindow:Layout(lanes)
	local w = self.size * self.lanes + 4;

	if not lanes then
		self:SetWidth(w);
		for k, v in pairs(self.regimes) do
			v:SetWidth(w);
		end
	else
		local h = self.size + self.seconds * 10 + 10;
		local m = math.max(self.settings.blinding, self.settings.test, self.settings.light);

		if self.regimes ~= nil then
			for k, v in pairs(self.regimes) do
				v:SetParent(nil);
			end
		end

		self.regimes = { };

		self:SetSize(w, h);

		self.regimes.recovery = CCTimerRegime("recovery state\non entry, it is safe to use fan the flames on stunned targets");
		self.regimes.recovery:SetParent(self);
		self.regimes.recovery:SetBackColor(Turbine.UI.Color(0.25, 1.0, 0.0, 0.0));
		self.regimes.recovery:SetSize(w, 10);
		self.regimes.recovery:SetPosition(0, h - 10);
		self.regimes.recovery.start = 0.0;
		self.regimes.recovery.stop = -1.0;
		self.regimes.recovery.color = Turbine.UI.Color(1.0, 1.0, 0.0, 0.0);

		self.regimes.reroot = CCTimerRegime("on entry, it is safe to use cracked earth on rooted targets");
		self.regimes.reroot:SetParent(self);
		self.regimes.reroot:SetBackColor(Turbine.UI.Color(0.25, 0.0, 0.0, 1.0));
		self.regimes.reroot:SetSize(w, (self.settings.reroot - m) * 10);
		self.regimes.reroot:SetPosition(0, h - self.settings.reroot * 10 - 10);
		self.regimes.reroot.start = self.settings.reroot;
		self.regimes.reroot.stop = m;
		self.regimes.reroot.color = Turbine.UI.Color(1.0, 0.0, 0.0, 1.0);

		local s = { };
		local e = { };
		local t = { };

		s.blinding = { };
		s.blinding.start = self.settings.blinding * 10;
		s.blinding.stop = self.settings.blinding * 10 - 11;
		s.blinding.r = 1.0;
		s.blinding.g = 1.0;
		s.blinding.b = 0.0;

		s.light = { };
		s.light.start = self.settings.light * 10;
		s.light.stop = self.settings.light * 10 - 14;
		s.light.r = 1.0;
		s.light.g = 0.0;
		s.light.b = 0.0;

		s.test = { };
		s.test.start = self.settings.test * 10;
		s.test.stop = self.settings.test * 10 - 11;
		s.test.r = 1.0;
		s.test.g = 0.0;
		s.test.b = 0.0;

		for k, v in pairs(s) do
			e[v.stop] = { n = 0, r = 0, g = 0, b = 0 };
		end

		for k, v in pairs(s) do
			if e[v.start] == nil then
				e[v.start] = { n = 0, r = 0, g = 0, b = 0 };
			end
			e[v.start].n = e[v.start].n + 1;
			e[v.start].r = e[v.start].r + v.r;
			e[v.start].g = e[v.start].g + v.g;
			e[v.start].b = e[v.start].b + v.b;

			if e[v.stop] == nil then
				e[v.stop] = { n = 0, r = 0, g = 0, b = 0 };
			end

			e[v.stop].n = e[v.stop].n - 1;
			e[v.stop].r = e[v.stop].r - v.r;
			e[v.stop].g = e[v.stop].g - v.g;
			e[v.stop].b = e[v.stop].b - v.b;
		end

		for k, v in pairs(e) do
			table.insert(t, k);
		end
		table.sort(t, function (a, b) return a > b end);

		local r = 0;
		local g = 0;
		local b = 0;
		local n = 0;
		local rc, gc, bc, mc;

		for k, v in ipairs(t) do
			n = n + e[v].n;
			r = r + e[v].r;
			g = g + e[v].g;
			b = b + e[v].b;

			if n > 0 then
				msg = "";
				if v == s.light.start then
					msg = msg .. "on entry, it is not safe to use light of the rising dawn on dazed targets\n";
					msg = msg .. "on entry, it is safe to use bane flare on stunned targets\n";
				end
				if v == s.test.start then
					msg = msg .. "on entry, it is not safe to use test of will on dazed targets\n";
				end
				if v == s.blinding.start then
					msg = msg .. "on entry, it is not safe to use blinding flash on dazed targets\n";
				end

				if t[k + 1] == s.light.stop then
					msg = msg .. "on exit, it is safe to use light of the rising dawn on dazed/stunned targets\n";
				end
				if t[k + 1] == s.test.stop then
					msg = msg .. "on exit, it is safe to use test of will on dazed/stunned targets\n";
				end
				if t[k + 1] == s.blinding.stop then
					msg = msg .. "on exit, it is safe to use blinding flash on dazed/stunned targets\n";
				end

				self.regimes[v] = CCTimerRegime(msg);
				self.regimes[v]:SetParent(self);
				rc = r / n;
				gc = g / n;
				bc = b / n;
				mc = math.max(rc, gc, bc);
				if n > 1 then
					rc = 1.0 - rc / mc;
					gc = 1.0 - gc / mc;
					bc = 1.0 - bc / mc;
					mc = math.max(rc, gc, bc);
				end
				if mc > 0 then
					self.regimes[v]:SetBackColor(Turbine.UI.Color(0.25, rc / mc, gc / mc, bc / mc));
					self.regimes[v].color = Turbine.UI.Color(1.0, rc /mc, gc / mc, bc / mc);
				else
					self.regimes[v]:SetBackColor(Turbine.UI.Color(0.25, 0.0, 0.0, 0.0));
					self.regimes[v].color = Turbine.UI.Color(1.0, 0.0, 0.0, 0.0);
				end
				self.regimes[v]:SetMouseVisible(false);
				self.regimes[v]:SetSize(w, v - t[k + 1]);
				self.regimes[v]:SetPosition(0, h - v - 10);
				self.regimes[v].start = v / 10;
				self.regimes[v].stop = t[k + 1] / 10;
			end
		end
	end
end

function CCTimerWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left and self.move then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.dragging = true;
	end
end

function CCTimerWindow:MouseMove(args)
	if self.dragging then
		local x, y = self:GetPosition();
		x = x + (args.X - self.dragStartX);
		y = y + (args.Y - self.dragStartY);
		self.settings.x = x;
		self.settings.y = y;
		self:SetPosition(x, y);
	end
end

function CCTimerWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self:SaveSettings();
		self.dragging = false;
	end
end

function CCTimerWindow:UpdateLayout()
	local time = Turbine.Engine.GetGameTime();

	self.track:set(1);
	for _, v in ipairs(self.cc) do
		if v.startTime > time then
			v.lane = nil;
		else
			local y = math.floor((v.endTime - time) * 10);
			v.lane = self.track:place(-y);
		end
	end

	if self.lanes ~= #self.track.track then
		self.lanes = #self.track.track;
		self:Layout(false);
		self:SetVisible(false);
		self:SetVisible(true);
	end
end

function CCTimerWindow:Update()
	local time = Turbine.Engine.GetGameTime();
	local delta;
	local changed = false;
	local t = { };

	for k, v in ipairs(self.cc) do
		delta = v.endTime - time;
		if delta < 0.0 and v.pre then
			changed = true;
			v.pre = false;
			v.endTime = v.endTime + self.duration[v.type];
			v.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, quickslot["herb"]));
		elseif delta < -1.0 or (delta < 0.0 and not recovery[v.type]) then
			changed = true;
			table.insert(t, 1, k);
			v.control:SetParent(nil);
		elseif v.startTime <= time and v.lane == nil then
			changed = true;
		end
	end

	for _, v in ipairs(t) do
		table.remove(self.cc, v);
	end

	if changed then
		table.sort(self.cc, function(a, b) return a.endTime < b.endTime; end);
		self:UpdateLayout();
	end

	for _, v in ipairs(self.cc) do
		if v.lane == nil then
			v.control:SetVisible(false);
		else
			local t = v.endTime - time;

			v.control:SetVisible(true);
			v.control:SetPosition(self.size * (v.lane - 1) + 2, (self.seconds - t) * 10);
			v.control:SetBackColor(Turbine.UI.Color(1.0, 0.0, 1.0, 0.0));
			for k1, v1 in pairs(self.regimes) do
				if t <= v1.start and t >= v1.stop then
					v.control:SetBackColor(v1.color);
					break;
				end
			end
		end
	end

	if #self.cc == 0 then
		self:SetWantsUpdates(false);
		self:SetVisible(false);
	end
end

function CCTimerWindow:AddCC(k, time)
	local item = { };

	item.control = Turbine.UI.Control();
	item.quickslot = Turbine.UI.Lotro.Quickslot();
	item.type = k;
	item.startTime = time + self.settings.cc[k];
	if k == "cracked" then
		item.endTime = item.startTime + self.preroot;
		item.pre = true;
	else
		item.endTime = item.startTime + self.duration[k];
		item.pre = false;
	end

	item.control:SetParent(self);
	item.control:SetSize(self.size, self.size);
	item.control:SetBackColor(Turbine.UI.Color(1.0, 0.0, 0.0, 0.0));
	item.control:SetZOrder(1);
	item.quickslot:SetParent(item.control);
	item.quickslot:SetPosition(-1, -1);
	item.quickslot:SetSize(self.size, self.size);
	item.quickslot:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, quickslot[k]));
	item.quickslot:SetMouseVisible(false);

	table.insert(self.cc, item);
	table.sort(self.cc, function(a, b) return a.endTime < b.endTime; end);

	self.lastTime = time;
	self:UpdateLayout();
	self:Update();
	self:SetVisible(true);
	self:SetWantsUpdates(true);
end

function CCTimerWindow:KeyDown(args)
	local time = Turbine.Engine.GetGameTime();

	if args.Action == 0x1000007B then
		if self.move then
			self.move = false;
			if #self.cc == 0 then
				self:SetVisible(false);
			end
			if self.dragging then
				self:SaveSettings();
				self.dragging = false;
			end
		else
			self.move = true;
			self:SetVisible(true);
		end

		self:SetMouseVisible(self.move);
		for k, v in pairs(self.regimes) do
			v:SetMouseVisible(self.move);
		end
	elseif args.Action == 0x100000B3 then
		self.hide = not self.hide;
		self:SetVisible();
	end
end

function CCTimerWindow:SaveSettings()
	Turbine.PluginData.Save(Turbine.DataScope.Character, "CCTimer", serialize(self.settings));
end

function CCTimerWindow:LoadSettings()
	self.settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "CCTimer");
	if type(settings) == "string" then
		settings = loadstring("return " .. settings .. ";")();
	end

	if type(self.settings) ~= "table" then
		self.settings = { };
	end

	if not self.settings.cc then
		self.settings.cc = { };
	end

	if not self.settings.cc.blinding then
		self.settings.cc.blinding = 1.1;
	end

	if not self.settings.cc.bane then
		self.settings.cc.bane = 1.1;
	end

	if not self.settings.cc.light then
		self.settings.cc.light = 0.2;
	end

	if not self.settings.cc.test then
		self.settings.cc.test = 0.7;
	end

	if not self.settings.cc.herb then
		self.settings.cc.herb = 0.0;
	end

	if not self.settings.cc.storm then
		self.settings.cc.storm = -0.2;
	end

	if not self.settings.cc.cracked then
		self.settings.cc.cracked = 0.1;
	end

	if not self.settings.cc.march then
		self.settings.cc.march = 1.1;
	end

	if not self.settings.x then
		self.settings.x = 0;
	end

	if not self.settings.y then
		self.settings.y = 0;
	end

	if not self.settings.blinding then
		self.settings.blinding = 1.5;
	end

	if not self.settings.light then
		self.settings.light = 1.8;
	end

	if not self.settings.test then
		self.settings.test = 3.1;
	end

	if not self.settings.reroot then
		self.settings.reroot = 7.3;
	end
end

CCTimerCommand = class(Turbine.ShellCommand);

function CCTimerCommand:Constructor(window)
	Turbine.ShellCommand.Constructor(self);

	self.window = window;
end

function CCTimerCommand:Execute(command, arguments)
	local args = snarfargs(arguments);

	local i = string.find(arguments, " ");
	local arg, cc, rest;

	if i ~= nil then
		arg = trim(string.sub(arguments, 0, i - 1));
		rest = trim(string.sub(arguments, i + 1, -1));
		if rest == "" then
			rest = nil;
		end
	else
		arg = arguments;
		rest = nil;
	end

	if args[1] == "timer" then
		local time = Turbine.Engine.GetGameTime();
		Turbine.Shell.WriteLine((time - self.window.lastTime) .. " seconds.");
	elseif args[1] == "reroot" then
		if tonumber(args[2]) ~= nil then
			self.window.settings.reroot = tonumber(args[2]);
			self.window:Layout(true);
			self.window:SaveSettings();
		end

		Turbine.Shell.WriteLine("CCTimer reroot time is " .. self.window.settings.reroot);
	elseif args[1] == "blinding" then
		if tonumber(args[2]) ~= nil then
			self.window.settings.blinding = tonumber(args[2]);
			self.window:Layout(true);
			self.window:SaveSettings();
		end

		Turbine.Shell.WriteLine("Blinding Flash unsafe time is " .. self.window.settings.blinding);
	elseif args[1] == "light" then
		if tonumber(args[2]) ~= nil then
			self.window.settings.light = tonumber(args[2]);
			self.window:Layout(true);
			self.window:SaveSettings();
		end

		Turbine.Shell.WriteLine("Light of the Rising Dawn unsafe time is " .. self.window.settings.light);
	elseif args[1] == "test" then
		if tonumber(args[2]) ~= nil then
			self.window.settings.test = tonumber(args[2]);
			self.window:Layout(true);
			self.window:SaveSettings();
		end

		Turbine.Shell.WriteLine("Test of Wills unsafe time is " .. self.window.settings.test);
	elseif args[1] == "act" and quickslot[args[2]] ~= nil then
		if tonumber(args[3]) ~= nil then
			self.window.settings.cc[args[2]] = tonumber(args[3]);
			self.window:SaveSettings();
		end

		if self.window.settings.cc[args[2]] ~= nil then
			Turbine.Shell.WriteLine("CCTimer " .. args[2] .. "(act) is " .. self.window.settings.cc[args[2]]);
		else
			Turbine.Shell.WriteLine("CCTimer " .. args[2] .. "(act) is 0");
		end
	else
		Turbine.Shell.WriteLine("usage: /cctimer <act <cc> <#>|reroot <#>|<blinding|light|test> <#>|timer>");
	end
end

function CCTimerCommand:GetShortHelp()
	return "Loremaster Crowd Control Timer";
end

plugin.command = CCTimerCommand(CCTimerWindow());
Turbine.Shell.AddCommand("cctimer", plugin.command);
Turbine.Shell.AddCommand("cc", plugin.command);